@@ -8,10 +8,11 @@
 use std::ops::Add;
 
 use chrono::{DateTime, Duration, Utc};
+use pgp::crypto::aead::AeadAlgorithm;
 use pgp::crypto::hash::HashAlgorithm;
 use pgp::crypto::sym::SymmetricKeyAlgorithm;
 use pgp::packet::{SubpacketData, SubpacketType};
-use pgp::types::{SignedUser, SignedUserAttribute, Tag};
+use pgp::types::{PublicKeyTrait, SignedUser, SignedUserAttribute, Tag};
 use pgp::{Signature, SignedKeyDetails, SignedPublicKey, SignedPublicSubKey};
 
 use crate::key::component::{ComponentKeyPub, SignedComponentKey, SignedComponentKeyPub};
@@ -104,10 +105,7 @@ impl CheckedCertificate {
         // Filter based on component key validity
         self.signing_capable_component_keys(reference)
             .filter(|sckp| sckp.is_component_subkey_valid_at(reference))
-            .filter(|sckp| {
-                let key_creation = sckp.key_creation_time();
-                sckp.has_valid_backsig_at(key_creation, reference)
-            })
+            .filter(|sckp| sckp.has_valid_backsig_at(reference))
             .filter(|sckp| policy::acceptable_pk_algorithm(sckp.public_params(), reference))
             .map(ComponentKeyPub::from)
             .map(Into::into)
@@ -338,6 +336,20 @@ impl CheckedCertificate {
         None
     }
 
+    pub fn preferred_aead_algo<'a>(
+        &'a self,
+        reference: &'a DateTime<Utc>,
+    ) -> Option<&'a [(SymmetricKeyAlgorithm, AeadAlgorithm)]> {
+        let prim_bind = self.primary_user_id_binding(reference);
+        if let Some(prim_bind) = prim_bind {
+            return Some(prim_bind.preferred_aead_algs());
+        }
+
+        // FIXME: handle dks
+
+        None
+    }
+
     pub fn preferred_hash_algorithms<'a>(
         &'a self,
         reference: &'a DateTime<Utc>,
@@ -351,6 +363,26 @@ impl CheckedCertificate {
 
         None
     }
+
+    pub fn features<'a>(&'a self, reference: &'a DateTime<Utc>) -> Option<u8> {
+        let prim_bind = self.primary_user_id_binding(reference);
+        if let Some(prim_bind) = prim_bind {
+            let feat = prim_bind.features();
+            if feat.len() == 1 {
+                return Some(feat[0]);
+            }
+        }
+
+        // FIXME: get dks at reference time
+        if let Some(dks) = self.dks() {
+            let feat = dks.features();
+            if feat.len() == 1 {
+                return Some(feat[0]);
+            }
+        }
+
+        None
+    }
 }
 
 impl From<&Certificate> for CheckedCertificate {
