@@ -6,18 +6,24 @@
 //! NOTE: This module is particularly experimental, the API may change drastically.
 //! (This current implementation does a lot of cloning and is limited to read-only operations.)
 
+use std::fmt::{Debug, Formatter};
+use std::io::Write;
+
 use chrono::{DateTime, Utc};
 use pgp::crypto::hash::HashAlgorithm;
 use pgp::crypto::public_key::PublicKeyAlgorithm;
 use pgp::crypto::sym::SymmetricKeyAlgorithm;
 use pgp::decrypt_session_key;
 use pgp::packet::{KeyFlags, PublicKeyEncryptedSessionKey, SubpacketData};
-use pgp::types::{KeyId, KeyTrait, PublicParams};
+use pgp::types::{
+    EskType, Fingerprint, KeyId, KeyVersion, PkeskBytes, PkeskVersion, PublicKeyTrait,
+    PublicParams, SignatureBytes,
+};
 use pgp::{
     packet, Message, Signature, SignedPublicKey, SignedPublicSubKey, SignedSecretKey,
     SignedSecretSubKey,
 };
-use rand::Rng;
+use rand::{thread_rng, Rng};
 use rand_core::CryptoRng;
 
 use crate::key::primary_user_id_binding_at;
@@ -33,6 +39,97 @@ pub enum SignedComponentKey {
     Sec(SignedComponentKeySec),
 }
 
+impl Debug for SignedComponentKey {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        match self {
+            SignedComponentKey::Pub(p) => p.fmt(f),
+            SignedComponentKey::Sec(s) => s.fmt(f),
+        }
+    }
+}
+
+impl PublicKeyTrait for SignedComponentKey {
+    fn version(&self) -> KeyVersion {
+        match self {
+            SignedComponentKey::Pub(p) => p.version(),
+            SignedComponentKey::Sec(s) => s.version(),
+        }
+    }
+
+    fn fingerprint(&self) -> Fingerprint {
+        match self {
+            SignedComponentKey::Pub(p) => p.fingerprint(),
+            SignedComponentKey::Sec(s) => s.fingerprint(),
+        }
+    }
+
+    fn key_id(&self) -> KeyId {
+        match self {
+            SignedComponentKey::Pub(p) => p.key_id(),
+            SignedComponentKey::Sec(s) => s.key_id(),
+        }
+    }
+
+    fn algorithm(&self) -> PublicKeyAlgorithm {
+        match self {
+            SignedComponentKey::Pub(p) => p.algorithm(),
+            SignedComponentKey::Sec(s) => s.algorithm(),
+        }
+    }
+
+    fn created_at(&self) -> &DateTime<Utc> {
+        match self {
+            SignedComponentKey::Pub(p) => p.created_at(),
+            SignedComponentKey::Sec(s) => s.created_at(),
+        }
+    }
+
+    fn expiration(&self) -> Option<u16> {
+        match self {
+            SignedComponentKey::Pub(p) => p.expiration(),
+            SignedComponentKey::Sec(s) => s.expiration(),
+        }
+    }
+
+    fn verify_signature(
+        &self,
+        hash: HashAlgorithm,
+        data: &[u8],
+        sig: &SignatureBytes,
+    ) -> pgp::errors::Result<()> {
+        match self {
+            SignedComponentKey::Pub(p) => p.verify_signature(hash, data, sig),
+            SignedComponentKey::Sec(s) => s.verify_signature(hash, data, sig),
+        }
+    }
+
+    fn encrypt<R: CryptoRng + Rng>(
+        &self,
+        rng: R,
+        plain: &[u8],
+        typ: EskType,
+    ) -> pgp::errors::Result<PkeskBytes> {
+        match self {
+            SignedComponentKey::Pub(p) => p.encrypt(rng, plain, typ),
+            SignedComponentKey::Sec(s) => s.encrypt(rng, plain, typ),
+        }
+    }
+
+    fn serialize_for_hashing(&self, writer: &mut impl Write) -> pgp::errors::Result<()> {
+        match self {
+            SignedComponentKey::Pub(p) => p.serialize_for_hashing(writer),
+            SignedComponentKey::Sec(s) => s.serialize_for_hashing(writer),
+        }
+    }
+
+    fn public_params(&self) -> &PublicParams {
+        match self {
+            SignedComponentKey::Pub(p) => p.public_params(),
+            SignedComponentKey::Sec(s) => s.public_params(),
+        }
+    }
+}
+
 enum SignedComponentPrimary {
     Pub(SignedPublicKey),
     Sec(SignedSecretKey),
@@ -119,87 +216,6 @@ impl From<&SignedComponentKey> for PrimaryOrSubkey {
 }
 
 impl SignedComponentKey {
-    pub fn version(&self) -> u8 {
-        match self {
-            SignedComponentKey::Pub(SignedComponentKeyPub::Primary((spk, _))) => {
-                spk.primary_key.version().into()
-            }
-            SignedComponentKey::Pub(SignedComponentKeyPub::Subkey((spsk, _))) => {
-                spsk.key.version().into()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Primary(ssk)) => {
-                ssk.primary_key.version().into()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Subkey((sssk, _))) => {
-                sssk.key.version().into()
-            }
-        }
-    }
-
-    pub fn public_params(&self) -> &PublicParams {
-        match self {
-            SignedComponentKey::Pub(SignedComponentKeyPub::Primary((spk, _))) => {
-                spk.primary_key.public_params()
-            }
-            SignedComponentKey::Pub(SignedComponentKeyPub::Subkey((spsk, _))) => {
-                spsk.key.public_params()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Primary(ssk)) => {
-                ssk.primary_key.public_params()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Subkey((sssk, _))) => {
-                sssk.key.public_params()
-            }
-        }
-    }
-
-    pub fn fingerprint(&self) -> Vec<u8> {
-        match self {
-            SignedComponentKey::Pub(SignedComponentKeyPub::Primary((spk, _))) => {
-                spk.primary_key.fingerprint()
-            }
-            SignedComponentKey::Pub(SignedComponentKeyPub::Subkey((spsk, _))) => {
-                spsk.key.fingerprint()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Primary(ssk)) => {
-                ssk.primary_key.fingerprint()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Subkey((sssk, _))) => {
-                sssk.key.fingerprint()
-            }
-        }
-    }
-
-    pub fn key_id(&self) -> KeyId {
-        match self {
-            SignedComponentKey::Pub(SignedComponentKeyPub::Primary((spk, _))) => {
-                spk.primary_key.key_id()
-            }
-            SignedComponentKey::Pub(SignedComponentKeyPub::Subkey((spsk, _))) => spsk.key.key_id(),
-            SignedComponentKey::Sec(SignedComponentKeySec::Primary(ssk)) => {
-                ssk.primary_key.key_id()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Subkey((sssk, _))) => sssk.key.key_id(),
-        }
-    }
-
-    pub fn created_at(&self) -> &DateTime<Utc> {
-        match self {
-            SignedComponentKey::Pub(SignedComponentKeyPub::Primary((spk, _))) => {
-                spk.primary_key.created_at()
-            }
-            SignedComponentKey::Pub(SignedComponentKeyPub::Subkey((spsk, _))) => {
-                spsk.key.created_at()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Primary(ssk)) => {
-                ssk.primary_key.created_at()
-            }
-            SignedComponentKey::Sec(SignedComponentKeySec::Subkey((sssk, _))) => {
-                sssk.key.created_at()
-            }
-        }
-    }
-
     pub fn sigs(&self) -> SigStack {
         match self {
             SignedComponentKey::Pub(SignedComponentKeyPub::Primary((_spk, sigs))) => {
@@ -209,7 +225,10 @@ impl SignedComponentKey {
                 SigStack::from_iter(spsk.signatures.iter())
             }
             SignedComponentKey::Sec(SignedComponentKeySec::Primary(_ssk)) => {
-                todo!() // FIXME
+                // FIXME: we don't want to support this kind of call at all (we're not handling DKS here)
+                // TSKs should probably be represented as Certificate + "secret key packets"
+
+                unimplemented!("unsupported case [primary secret]")
             }
             SignedComponentKey::Sec(SignedComponentKeySec::Subkey((sssk, _))) => {
                 SigStack::from_iter(sssk.signatures.iter())
@@ -340,6 +359,88 @@ pub enum SignedComponentKeyPub {
     Subkey((SignedPublicSubKey, Option<Signature>)),
 }
 
+impl PublicKeyTrait for SignedComponentKeyPub {
+    fn version(&self) -> KeyVersion {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.version(),
+            SignedComponentKeyPub::Subkey((s, _)) => s.version(),
+        }
+    }
+
+    fn fingerprint(&self) -> Fingerprint {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.fingerprint(),
+            SignedComponentKeyPub::Subkey((s, _)) => s.fingerprint(),
+        }
+    }
+
+    fn key_id(&self) -> KeyId {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.key_id(),
+            SignedComponentKeyPub::Subkey((s, _)) => s.key_id(),
+        }
+    }
+
+    fn algorithm(&self) -> PublicKeyAlgorithm {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.algorithm(),
+            SignedComponentKeyPub::Subkey((s, _)) => s.algorithm(),
+        }
+    }
+
+    fn created_at(&self) -> &DateTime<Utc> {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.created_at(),
+            SignedComponentKeyPub::Subkey((s, _)) => s.created_at(),
+        }
+    }
+
+    fn expiration(&self) -> Option<u16> {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.expiration(),
+            SignedComponentKeyPub::Subkey((s, _)) => s.expiration(),
+        }
+    }
+
+    fn verify_signature(
+        &self,
+        hash: HashAlgorithm,
+        data: &[u8],
+        sig: &SignatureBytes,
+    ) -> pgp::errors::Result<()> {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.verify_signature(hash, data, sig),
+            SignedComponentKeyPub::Subkey((s, _)) => s.verify_signature(hash, data, sig),
+        }
+    }
+
+    fn encrypt<R: CryptoRng + Rng>(
+        &self,
+        rng: R,
+        plain: &[u8],
+        typ: EskType,
+    ) -> pgp::errors::Result<PkeskBytes> {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.encrypt(rng, plain, typ),
+            SignedComponentKeyPub::Subkey((s, _)) => s.encrypt(rng, plain, typ),
+        }
+    }
+
+    fn serialize_for_hashing(&self, writer: &mut impl Write) -> pgp::errors::Result<()> {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.serialize_for_hashing(writer),
+            SignedComponentKeyPub::Subkey((s, _)) => s.serialize_for_hashing(writer),
+        }
+    }
+
+    fn public_params(&self) -> &PublicParams {
+        match self {
+            SignedComponentKeyPub::Primary((p, _)) => p.public_params(),
+            SignedComponentKeyPub::Subkey((s, _)) => s.public_params(),
+        }
+    }
+}
+
 impl From<SignedComponentKeyPub> for ComponentKeyPub {
     fn from(value: SignedComponentKeyPub) -> ComponentKeyPub {
         match value {
@@ -368,8 +469,12 @@ impl ComponentKeySec {
         F: FnOnce() -> String,
     {
         match self {
-            ComponentKeySec::Primary(sk) => Ok(msg.sign(&sk, key_pw, hash_algo)?),
-            ComponentKeySec::Subkey(ssk) => Ok(msg.sign(&ssk, key_pw, hash_algo)?),
+            ComponentKeySec::Primary(sk) => {
+                Ok(msg.sign(&mut thread_rng(), &sk, key_pw, hash_algo)?)
+            }
+            ComponentKeySec::Subkey(ssk) => {
+                Ok(msg.sign(&mut thread_rng(), &ssk, key_pw, hash_algo)?)
+            }
         }
     }
 
@@ -380,7 +485,7 @@ impl ComponentKeySec {
         }
     }
 
-    pub fn fingerprint(&self) -> Vec<u8> {
+    pub fn fingerprint(&self) -> Fingerprint {
         match self {
             ComponentKeySec::Primary(pri) => pri.fingerprint(),
             ComponentKeySec::Subkey(sub) => sub.fingerprint(),
@@ -418,6 +523,97 @@ pub enum SignedComponentKeySec {
     Subkey((SignedSecretSubKey, Option<Signature>)), // key, dks
 }
 
+impl Debug for SignedComponentKeySec {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.fmt(f),
+            SignedComponentKeySec::Subkey(s) => s.fmt(f),
+        }
+    }
+}
+
+impl PublicKeyTrait for SignedComponentKeySec {
+    fn version(&self) -> KeyVersion {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.version(),
+            SignedComponentKeySec::Subkey((s, _)) => s.version(),
+        }
+    }
+
+    fn fingerprint(&self) -> Fingerprint {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.fingerprint(),
+            SignedComponentKeySec::Subkey((s, _)) => s.fingerprint(),
+        }
+    }
+
+    fn key_id(&self) -> KeyId {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.key_id(),
+            SignedComponentKeySec::Subkey((s, _)) => s.key_id(),
+        }
+    }
+
+    fn algorithm(&self) -> PublicKeyAlgorithm {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.algorithm(),
+            SignedComponentKeySec::Subkey((s, _)) => s.algorithm(),
+        }
+    }
+
+    fn created_at(&self) -> &DateTime<Utc> {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.created_at(),
+            SignedComponentKeySec::Subkey((s, _)) => s.created_at(),
+        }
+    }
+
+    fn expiration(&self) -> Option<u16> {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.expiration(),
+            SignedComponentKeySec::Subkey((s, _)) => s.expiration(),
+        }
+    }
+
+    fn verify_signature(
+        &self,
+        hash: HashAlgorithm,
+        data: &[u8],
+        sig: &SignatureBytes,
+    ) -> pgp::errors::Result<()> {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.verify_signature(hash, data, sig),
+            SignedComponentKeySec::Subkey((s, _)) => s.verify_signature(hash, data, sig),
+        }
+    }
+
+    fn encrypt<R: CryptoRng + Rng>(
+        &self,
+        rng: R,
+        plain: &[u8],
+        typ: EskType,
+    ) -> pgp::errors::Result<PkeskBytes> {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.encrypt(rng, plain, typ),
+            SignedComponentKeySec::Subkey((s, _)) => s.encrypt(rng, plain, typ),
+        }
+    }
+
+    fn serialize_for_hashing(&self, writer: &mut impl Write) -> pgp::errors::Result<()> {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.serialize_for_hashing(writer),
+            SignedComponentKeySec::Subkey((s, _)) => s.serialize_for_hashing(writer),
+        }
+    }
+
+    fn public_params(&self) -> &PublicParams {
+        match self {
+            SignedComponentKeySec::Primary(p) => p.public_params(),
+            SignedComponentKeySec::Subkey((s, _)) => s.public_params(),
+        }
+    }
+}
+
 pub(crate) enum KeyFlagMatch {
     Sign,
     Enc,
@@ -472,12 +668,23 @@ impl SignedComponentKeySec {
     where
         F: FnOnce() -> String,
     {
+        let typ = match pkesk.version() {
+            PkeskVersion::V3 => pgp::types::EskType::V3_4,
+            PkeskVersion::V6 => pgp::types::EskType::V6,
+            v => {
+                return Err(Error::Rpgp(pgp::errors::Error::Message(format!(
+                    "Unsupported PKESK version {:?}",
+                    v
+                ))))
+            }
+        };
+
         match self {
             SignedComponentKeySec::Primary(ssk) => {
-                Ok(decrypt_session_key(&ssk, key_pw, pkesk.mpis())?)
+                Ok(decrypt_session_key(&ssk, key_pw, pkesk.values()?, typ)?)
             }
             SignedComponentKeySec::Subkey((sssk, _)) => {
-                Ok(decrypt_session_key(&sssk, key_pw, pkesk.mpis())?)
+                Ok(decrypt_session_key(&sssk, key_pw, pkesk.values()?, typ)?)
             }
         }
     }
@@ -513,15 +720,13 @@ impl SignedComponentKeyPub {
 
     /// This fn is intended for use on signing-capable component keys.
     /// It returns "false" is a subkey is not validly "backward-bound" to the primary.
-    pub(crate) fn has_valid_backsig_at(
-        &self,
-        key_creation: &DateTime<Utc>,
-        reference: &DateTime<Utc>,
-    ) -> bool {
+    pub(crate) fn has_valid_backsig_at(&self, reference: &DateTime<Utc>) -> bool {
         if let SignedComponentKeyPub::Subkey(sssk) = self {
             // If a subkey binding signature has no embedded back signature,
             // the subkey is not reasonably bound to the certificate for signing
 
+            let key_creation: &DateTime<Utc> = self.key_creation_time();
+
             let binding = SigStack::from_iter(sssk.0.signatures.iter()).active_at(Some(reference));
 
             if let Some(binding) = binding {
@@ -586,20 +791,20 @@ impl SignedComponentKeyPub {
 }
 
 impl ComponentKeyPub {
-    pub fn pkesk_from_session_key<R: CryptoRng + Rng>(
+    pub fn pkesk_from_session_key_v3<R: CryptoRng + Rng>(
         &self,
         rng: &mut R,
         session_key: &[u8],
         alg: SymmetricKeyAlgorithm,
     ) -> Result<PublicKeyEncryptedSessionKey, Error> {
         match &self {
-            Self::Primary(pk) => Ok(PublicKeyEncryptedSessionKey::from_session_key(
+            Self::Primary(pk) => Ok(PublicKeyEncryptedSessionKey::from_session_key_v3(
                 rng,
                 session_key,
                 alg,
                 pk,
             )?),
-            Self::Subkey(psk) => Ok(PublicKeyEncryptedSessionKey::from_session_key(
+            Self::Subkey(psk) => Ok(PublicKeyEncryptedSessionKey::from_session_key_v3(
                 rng,
                 session_key,
                 alg,
@@ -608,6 +813,25 @@ impl ComponentKeyPub {
         }
     }
 
+    pub fn pkesk_from_session_key_v6<R: CryptoRng + Rng>(
+        &self,
+        rng: &mut R,
+        session_key: &[u8],
+    ) -> Result<PublicKeyEncryptedSessionKey, Error> {
+        match &self {
+            Self::Primary(pk) => Ok(PublicKeyEncryptedSessionKey::from_session_key_v6(
+                rng,
+                session_key,
+                pk,
+            )?),
+            Self::Subkey(psk) => Ok(PublicKeyEncryptedSessionKey::from_session_key_v6(
+                rng,
+                session_key,
+                psk,
+            )?),
+        }
+    }
+
     pub fn verify(&self, s: &Signature, payload: &[u8]) -> Result<(), Error> {
         // policy check on s
         if !sig::signature_acceptable(s) {
@@ -636,7 +860,7 @@ impl ComponentKeyPub {
         }
     }
 
-    pub fn fingerprint(&self) -> Vec<u8> {
+    pub fn fingerprint(&self) -> Fingerprint {
         match self {
             ComponentKeyPub::Primary(pri) => pri.fingerprint(),
             ComponentKeyPub::Subkey(sub) => sub.fingerprint(),
