@@ -5,16 +5,18 @@
 
 use std::io;
 
-use chrono::{DateTime, Utc};
+use chrono::{DateTime, SubsecRound, Utc};
 use openpgp_card_rpgp::CardSlot;
 use pgp::cleartext::CleartextSignedMessage;
 use pgp::crypto::hash::HashAlgorithm;
+use pgp::packet::{SignatureConfig, SignatureType, Subpacket, SubpacketData};
 use pgp::ser::Serialize;
-use pgp::types::{KeyTrait, SignedUser};
+use pgp::types::{Fingerprint, KeyVersion, PublicKeyTrait, SignedUser};
 use pgp::{
     ArmorOptions, KeyType, Message, PublicOrSecret, SecretKeyParamsBuilder, Signature,
     SignedKeyDetails, SignedPublicKey, SignedSecretKey, StandaloneSignature, SubkeyParamsBuilder,
 };
+use rand::thread_rng;
 
 use crate::card::{card_by_pp, verify_pin_from_card_state};
 use crate::key::component::{
@@ -22,7 +24,8 @@ use crate::key::component::{
     SignedComponentKeySec,
 };
 use crate::policy::{
-    PREFERRED_COMPRESSION_ALGORITHMS, PREFERRED_HASH_ALGORITHMS, PREFERRED_SYMMETRIC_KEY_ALGORITHMS,
+    PREFERRED_AEAD_ALGORITHMS, PREFERRED_COMPRESSION_ALGORITHMS, PREFERRED_HASH_ALGORITHMS,
+    PREFERRED_SYMMETRIC_KEY_ALGORITHMS,
 };
 use crate::sig::stack::SigStack;
 use crate::{card, policy, Error};
@@ -139,7 +142,7 @@ pub(crate) fn primary_user_id_binding_at<'a>(
 }
 
 impl Certificate {
-    pub fn fingerprint(&self) -> Vec<u8> {
+    pub fn fingerprint(&self) -> Fingerprint {
         self.spk.primary_key.fingerprint()
     }
 
@@ -309,7 +312,12 @@ impl DataSigner {
                         &|| {},
                     )?;
 
-                    Ok(CleartextSignedMessage::sign(body, &cs, String::default)?)
+                    Ok(CleartextSignedMessage::sign(
+                        thread_rng(),
+                        body,
+                        &cs,
+                        String::default,
+                    )?)
                 } else {
                     Err(Error::Message("Card not found".to_string()))
                 }
@@ -319,7 +327,7 @@ impl DataSigner {
                 ComponentKeySec::Primary(sk) => key_pws
                     .iter()
                     .flat_map(|key_pw| {
-                        CleartextSignedMessage::sign(body, &sk, || {
+                        CleartextSignedMessage::sign(thread_rng(), body, &sk, || {
                             String::from_utf8_lossy(key_pw).to_string()
                         })
                     })
@@ -327,7 +335,7 @@ impl DataSigner {
                 ComponentKeySec::Subkey(ssk) => key_pws
                     .iter()
                     .flat_map(|key_pw| {
-                        CleartextSignedMessage::sign(body, &ssk, || {
+                        CleartextSignedMessage::sign(thread_rng(), body, &ssk, || {
                             String::from_utf8_lossy(key_pw).to_string()
                         })
                     })
@@ -339,7 +347,7 @@ impl DataSigner {
         }
     }
 
-    pub fn fingerprint(&self) -> Vec<u8> {
+    pub fn fingerprint(&self) -> Fingerprint {
         match self {
             DataSigner::Card(p) => p.fingerprint(),
             DataSigner::Software(s) => match s {
@@ -386,37 +394,49 @@ impl Tsk {
             ));
         }
 
-        let signed_secret_key: &Tsk = &tsks[0];
+        let tsk: &Tsk = &tsks[0];
+
+        let ssk = tsk.key();
 
         if armored {
-            signed_secret_key
-                .key()
-                .to_armored_writer(&mut sink, ArmorOptions::default())?;
+            ssk.to_armored_writer(&mut sink, ArmorOptions::default())?;
         } else {
-            signed_secret_key.key().to_writer(&mut sink)?;
+            ssk.to_writer(&mut sink)?;
         }
 
         Ok(())
     }
 
-    // FIXME: remove this from the API again?
+    /// This is a horrible accessor, and will be simplified once card handling is factored out
+    /// of rpgpie again.
+    ///
+    /// NOTE: panics if the Tsk is card-backed!
+    ///
+    /// FIXME: remove this from the API again?
     pub fn key(&self) -> &SignedSecretKey {
         match self {
             Tsk::Tsk(ssk) => ssk,
-            Tsk::Card(_) => panic!("card mode Tsk, only usable for HSM operations"), // FIXME
+            Tsk::Card(_) => {
+                unimplemented!("card mode Tsk, only usable for HSM operations")
+            }
         }
     }
 
-    pub fn generate(
+    pub fn generate_v4(
         key_type_primary: KeyType,
         key_type_encrypt: impl Into<Option<KeyType>>,
         primary_user_id: Option<String>,
         secondary_user_ids: Vec<String>,
+        key_password: Option<&str>,
     ) -> Result<Self, Error> {
         let Some(primary_user_id) = primary_user_id else {
-            panic!("Generating UID-less keys not supported");
+            return Err(Error::Rpgp(pgp::errors::Error::Message(
+                "Generating UID-less keys not supported".to_string(),
+            )));
         };
 
+        let mut rng = thread_rng();
+
         let subkeys = if let Some(key_type_encrypt) = key_type_encrypt.into() {
             vec![SubkeyParamsBuilder::default()
                 .key_type(key_type_encrypt)
@@ -439,10 +459,110 @@ impl Tsk {
             .subkeys(subkeys);
 
         let secret_key_params = key_params.build()?;
-        let secret_key = secret_key_params.generate()?;
+        let secret_key = secret_key_params.generate(&mut rng)?;
 
         let passwd_fn = String::new;
-        let signed_secret_key = secret_key.sign(passwd_fn)?;
+
+        let mut signed_secret_key = secret_key.sign(&mut thread_rng(), passwd_fn)?;
+        if let Some(key_password) = key_password {
+            signed_secret_key
+                .primary_key
+                .set_password(&mut rng, || key_password.to_string())?;
+
+            for sk in &mut signed_secret_key.secret_subkeys {
+                sk.key.set_password(&mut rng, || key_password.to_string())?;
+            }
+        };
+
+        Ok(Tsk::Tsk(signed_secret_key))
+    }
+
+    pub fn generate_v6(
+        key_type_primary: KeyType,
+        key_type_encrypt: KeyType,
+        primary_user_id: Option<String>,
+        secondary_user_ids: Vec<String>,
+        key_password: Option<&str>,
+    ) -> Result<Self, Error> {
+        let mut rng = thread_rng();
+
+        let (primary_user_id, uidless) = match primary_user_id {
+            Some(uid) => (uid, false),
+            None => (String::default(), true),
+        };
+
+        let mut key_params = SecretKeyParamsBuilder::default();
+        key_params
+            .version(KeyVersion::V6)
+            .key_type(key_type_primary)
+            .can_certify(true)
+            .can_sign(true)
+            .primary_user_id(primary_user_id)
+            .user_ids(secondary_user_ids)
+            .preferred_symmetric_algorithms(PREFERRED_SYMMETRIC_KEY_ALGORITHMS.into())
+            .preferred_hash_algorithms(PREFERRED_HASH_ALGORITHMS.into())
+            .preferred_compression_algorithms(PREFERRED_COMPRESSION_ALGORITHMS.into())
+            .preferred_aead_algorithms(PREFERRED_AEAD_ALGORITHMS.into())
+            .subkeys(vec![SubkeyParamsBuilder::default()
+                .version(KeyVersion::V6)
+                .key_type(key_type_encrypt)
+                .can_encrypt(true)
+                .build()?]);
+
+        let secret_key_params = key_params.build()?;
+        let secret_key = secret_key_params.generate(&mut rng)?;
+
+        let passwd_fn = String::new;
+        let mut signed_secret_key = secret_key.sign(&mut thread_rng(), passwd_fn)?;
+
+        // drop userid if it's just a placeholder
+        if uidless {
+            signed_secret_key.details.users = vec![];
+        }
+
+        // Add a DKS for v6 keys
+        let hash_alg = HashAlgorithm::SHA2_512; // FIXME?
+
+        let mut config = SignatureConfig::v6(
+            &mut thread_rng(),
+            SignatureType::Key,
+            signed_secret_key.algorithm(),
+            hash_alg,
+        )?;
+        config.hashed_subpackets = vec![
+            Subpacket::regular(SubpacketData::SignatureCreationTime(
+                chrono::Utc::now().trunc_subsecs(0),
+            )),
+            Subpacket::regular(SubpacketData::KeyFlags([3].into())), // FIXME, set CS
+            Subpacket::regular(SubpacketData::Features([0x01 | 0x08].into())), // FIXME, set SEIPDv2 + SEIPDv1
+            Subpacket::regular(SubpacketData::PreferredSymmetricAlgorithms(
+                PREFERRED_SYMMETRIC_KEY_ALGORITHMS.into(),
+            )),
+            Subpacket::regular(SubpacketData::PreferredHashAlgorithms(
+                PREFERRED_HASH_ALGORITHMS.into(),
+            )),
+            Subpacket::regular(SubpacketData::PreferredCompressionAlgorithms(
+                PREFERRED_COMPRESSION_ALGORITHMS.into(),
+            )),
+            Subpacket::regular(SubpacketData::IssuerFingerprint(
+                signed_secret_key.fingerprint(),
+            )),
+        ];
+
+        let passwd_fn = String::new;
+        let dks = config.sign_key(&signed_secret_key, passwd_fn, &signed_secret_key)?;
+
+        signed_secret_key.details.direct_signatures.push(dks);
+
+        if let Some(key_password) = key_password {
+            signed_secret_key
+                .primary_key
+                .set_password(&mut rng, || key_password.to_string())?;
+
+            for sk in &mut signed_secret_key.secret_subkeys {
+                sk.key.set_password(&mut rng, || key_password.to_string())?;
+            }
+        };
 
         Ok(Tsk::Tsk(signed_secret_key))
     }
