@@ -7,14 +7,16 @@ use std::io;
 use std::io::Read;
 
 use openpgp_card_rpgp::CardSlot;
+use pgp::crypto::aead::AeadAlgorithm;
 use pgp::crypto::hash::HashAlgorithm;
 use pgp::crypto::sym::SymmetricKeyAlgorithm;
 use pgp::packet::{LiteralData, Packet, SymEncryptedProtectedData, SymKeyEncryptedSessionKey};
 use pgp::ser::Serialize;
-use pgp::types::{KeyId, SecretKeyTrait, StringToKey, Tag};
+use pgp::types::{SecretKeyTrait, StringToKey};
 use pgp::{decrypt_session_key_with_password, ArmorOptions, PlainSessionKey};
 use pgp::{Deserializable, Edata, Esk};
 use pgp::{Message, Signature};
+use zeroize::Zeroizing;
 
 use crate::card::{card_by_pp, verify_pin_from_card_state};
 use crate::key::checked::CheckedCertificate;
@@ -74,12 +76,12 @@ pub fn unpack(
     skesk_passwords: Vec<&[u8]>,
     verifier: &[Certificate],
 ) -> Result<MessageResult, Error> {
-    unpack_int(msg, decryptor, key_passwords, skesk_passwords, verifier, 0)
+    unpack_int(&msg, decryptor, key_passwords, skesk_passwords, verifier, 0)
 }
 
 // internal variant of the unpack() function with depth handling and limitation
 fn unpack_int(
-    msg: Message,
+    msg: &Message,
     decryptors: &[Tsk],
     key_passwords: Vec<&[u8]>,
     skesk_passwords: Vec<&[u8]>,
@@ -99,9 +101,6 @@ fn unpack_int(
             'esks: for e in esk {
                 match e {
                     Esk::PublicKeyEncryptedSessionKey(pkesk) => {
-                        let encrypted_to = pkesk.id();
-                        let wildcard_id = KeyId::from_slice(&[0, 0, 0, 0, 0, 0, 0, 0])?;
-
                         if !decryptors.is_empty() {
                             let key_pw = match key_passwords.len() {
                                 0 => "".to_string(),
@@ -117,17 +116,18 @@ fn unpack_int(
 
                             for dec in decryptors {
                                 for ek in dec.decryption_capable_component_keys() {
-                                    // Match key id / or check for wildcard.
-                                    // FIXME: do we want any other checks?
-                                    if encrypted_to == &wildcard_id || encrypted_to == &ek.key_id()
-                                    {
+                                    // Match key id/fp or check for wildcard.
+                                    if pkesk.match_identity(&ek) {
                                         match ek {
                                             SignedComponentKey::Sec(s) => {
-                                                if let Ok(sk0) =
-                                                    s.decrypt_session_key(pkesk, || key_pw.clone())
+                                                match s
+                                                    .decrypt_session_key(pkesk, || key_pw.clone())
                                                 {
-                                                    sk = Some(sk0);
-                                                    break 'esks;
+                                                    Ok(sk0) => {
+                                                        sk = Some(sk0);
+                                                        break 'esks;
+                                                    }
+                                                    Err(e) => eprintln!("err {:?}", e),
                                                 }
                                             }
 
@@ -146,23 +146,13 @@ fn unpack_int(
                                                         &|| {}, // FIXME: touch prompt?
                                                     )?;
 
-                                                    let mpis =
-                                                        if let Esk::PublicKeyEncryptedSessionKey(
-                                                            ref k,
-                                                        ) = esk[0]
-                                                        {
-                                                            k.mpis()
-                                                        } else {
-                                                            panic!("whoops")
-                                                        };
-
                                                     if let Ok((
                                                         session_key,
                                                         session_key_algorithm,
                                                     )) = cs.unlock(String::new, |priv_key| {
-                                                        priv_key.decrypt(mpis)
+                                                        priv_key.decrypt(pkesk.values()?)
                                                     }) {
-                                                        sk = Some(PlainSessionKey::V4 {
+                                                        sk = Some(PlainSessionKey::V3_4 {
                                                             key: session_key,
                                                             sym_alg: session_key_algorithm,
                                                         });
@@ -200,8 +190,6 @@ fn unpack_int(
                 }
             }
 
-            let mut ciphertext = edata.data()[..].to_vec();
-
             let Some(sk) = sk else {
                 eprintln!("Failed to get session key");
 
@@ -209,79 +197,49 @@ fn unpack_int(
                 return Err(Error::Message("Failed to get session key".to_string()));
             };
 
-            let (decrypted_packet, alg, key) = match sk {
-                PlainSessionKey::V4 { sym_alg, key } => {
-                    let protected = edata.tag() == Tag::SymEncryptedProtectedData;
+            let (sym_alg, key) = match &sk {
+                PlainSessionKey::V3_4 { sym_alg, key } => (*sym_alg, key.to_vec()),
 
-                    let decrypted_packet: Vec<u8> = if protected {
-                        sym_alg
-                            .decrypt_protected(&key, &mut ciphertext)
-                            .expect("FIXME")
-                            .to_vec()
-                    } else {
-                        eprintln!("WARN: decrypting unprotected!");
-                        sym_alg
-                            .decrypt(&key, &mut ciphertext)
-                            .expect("FIXME")
-                            .to_vec()
-                    };
-
-                    (decrypted_packet, sym_alg, key)
-                }
-
-                PlainSessionKey::V6 { key } => {
-                    if let Edata::SymEncryptedProtectedData(seipd) = edata {
-                        if let pgp::packet::Data::V2 { sym_alg, .. } = &seipd.data() {
-                            let decrypted_packet =
-                                seipd.decrypt(&key, Some(*sym_alg)).expect("FIXME");
-
-                            (decrypted_packet, *sym_alg, key)
-                        } else {
-                            unimplemented!()
+                PlainSessionKey::V6 { key } => match edata {
+                    Edata::SymEncryptedProtectedData(seipd) => match &seipd.data() {
+                        pgp::packet::Data::V2 { sym_alg, .. } => (*sym_alg, key.to_vec()),
+                        pgp::packet::Data::V1 { .. } => {
+                            return Err(Error::Rpgp(pgp::errors::Error::Message(format!(
+                                "V6 session key with unexpected symmetric data version {}",
+                                seipd.version()
+                            ))));
                         }
-                    } else {
-                        unimplemented!()
+                    },
+                    Edata::SymEncryptedData(_) => {
+                        return Err(Error::Rpgp(pgp::errors::Error::Message(
+                            "SED packet is not supported with V6 session keys".to_string(),
+                        )));
                     }
-                }
+                },
 
-                _ => unimplemented!(),
+                psk => {
+                    return Err(Error::Rpgp(pgp::errors::Error::Message(format!(
+                        "Unsupported plain session key version {:?}",
+                        psk
+                    ))));
+                }
             };
 
-            // Store decrypted message, to be able to continue consuming the decryption output further
-            let mut mr: Option<MessageResult> = None;
+            let msg = edata.decrypt(sk.clone())?;
 
-            for msg in Message::from_bytes_many(decrypted_packet.as_slice()) {
-                let msg = msg?;
+            let mut mr = unpack_int(
+                &msg,
+                decryptors,
+                key_passwords.clone(),
+                skesk_passwords.clone(),
+                verifier,
+                depth + 1,
+            )?;
 
-                let res = unpack_int(
-                    msg,
-                    decryptors,
-                    key_passwords.clone(),
-                    skesk_passwords.clone(),
-                    verifier,
-                    depth + 1,
-                );
+            // Return the session key in MessageResult
+            mr.session_key = Some(((sym_alg).into(), key));
 
-                if mr.is_none() {
-                    mr = Some(res?);
-                } else {
-                    // A decrypted message should not contain *two* output messages, this is an error case
-                    return Err(Error::Message(
-                        "Got two messages from decryption process".to_string(),
-                    ));
-                }
-            }
-
-            if let Some(mut mr) = mr {
-                // Return the session key in MessageResult
-                mr.session_key = Some((alg.into(), key));
-
-                Ok(mr)
-            } else {
-                Err(Error::Message(
-                    "Got no message from decryption process".to_string(),
-                ))
-            }
+            Ok(mr)
         }
         Message::Literal(lit) => Ok(MessageResult {
             validated: vec![],
@@ -293,7 +251,7 @@ fn unpack_int(
             let msg = Message::from_bytes(payload)?;
 
             unpack_int(
-                msg,
+                &msg,
                 decryptors,
                 key_passwords,
                 skesk_passwords,
@@ -330,16 +288,22 @@ fn unpack_int(
                             grab_data(&msg, depth + 1)
                         }
                         Message::Signed { message, .. } => {
+                            let Some(message) = message else {
+                                return Err(Error::Message(
+                                    "No inner message found in signed message".to_string(),
+                                ));
+                            };
                             // FIXME: return raw message data, if notarizing signature?
-
-                            grab_data(message.as_ref().unwrap(), depth + 1)
+                            grab_data(message.as_ref(), depth + 1)
                         }
                         Message::Encrypted { .. } => Ok(m.to_bytes()?),
                     }
                 }
 
-                // FIXME: don't clone so much
-                let data: Vec<u8> = grab_data(inner.as_ref().unwrap(), 0)?;
+                let Some(inner) = inner else {
+                    return Err(Error::Message("Inner message is None".to_string()));
+                };
+                let data = grab_data(inner.as_ref(), 0)?;
 
                 for vcert in verifier {
                     if let Some(reference) = signature.created() {
@@ -356,11 +320,11 @@ fn unpack_int(
                 }
             }
 
-            // FIXME: don't clone
-            let inner = *inner.clone().expect("FIXME");
-
+            let Some(inner) = inner else {
+                return Err(Error::Message("Inner message is None".to_string()));
+            };
             let mut mr = unpack_int(
-                inner,
+                inner.as_ref(),
                 decryptors,
                 key_passwords,
                 skesk_passwords,
@@ -377,31 +341,51 @@ fn unpack_int(
     }
 }
 
+#[derive(Debug, Copy, Clone, PartialEq)]
+pub enum EncryptionMechanism {
+    SeipdV1(SymmetricKeyAlgorithm),
+    SeipdV2(AeadAlgorithm, SymmetricKeyAlgorithm),
+}
+
 /// Encrypt (and optionally sign) a message.
 ///
-/// NOTE: `source` is expected to contain raw data, not a Message
-// FIXME: clean up this API?
+/// NOTE: `source` is expected to contain raw data, not an OpenPGP Message
+///
+/// FIXME: we need passwords to unlock signers!
+///
+/// FIXME: pass recipient primary (to set as intended recipient)
 #[allow(clippy::too_many_arguments)]
 pub fn encrypt(
+    mechanism: EncryptionMechanism,
     recipients: Vec<ComponentKeyPub>,
     skesk_passwords: Vec<&[u8]>,
-    signers: Vec<Tsk>, // FIXME: passwords?
-    hash_algo: Option<&HashAlgorithm>,
-    symmetric_algo: SymmetricKeyAlgorithm,
+    signers: Vec<Tsk>,
+    hash_algo: Option<HashAlgorithm>,
     source: &mut (dyn Read + Send + Sync),
     mut sink: &mut (dyn io::Write + Send + Sync),
     armor: bool,
-) -> Result<Vec<u8>, Error> {
+) -> Result<Zeroizing<Vec<u8>>, Error> {
     let mut rng = rand::thread_rng();
 
     // 1. Define a new session key
-    let session_key = symmetric_algo.new_session_key(&mut rng);
+    let session_key = match mechanism {
+        EncryptionMechanism::SeipdV1(sym) | EncryptionMechanism::SeipdV2(_, sym) => {
+            sym.new_session_key(&mut rng)
+        }
+    };
 
     let mut esk = vec![];
 
     // 2. Encrypt (pub) the session key, to each PublicKey.
     for recipient in recipients {
-        let pkesk = recipient.pkesk_from_session_key(&mut rng, &session_key, symmetric_algo)?;
+        let pkesk = match mechanism {
+            EncryptionMechanism::SeipdV1(sym) => {
+                recipient.pkesk_from_session_key_v3(&mut rng, &session_key, sym)?
+            }
+            EncryptionMechanism::SeipdV2(_, _) => {
+                recipient.pkesk_from_session_key_v6(&mut rng, &session_key)?
+            }
+        };
 
         esk.push(Esk::PublicKeyEncryptedSessionKey(pkesk));
     }
@@ -410,12 +394,31 @@ pub fn encrypt(
     for pw in skesk_passwords {
         let pass = String::from_utf8_lossy(pw).to_string();
 
-        let skesk = SymKeyEncryptedSessionKey::encrypt(
-            || pass,
-            &session_key,
-            StringToKey::new_default(&mut rng),
-            symmetric_algo,
-        )?;
+        let skesk = match mechanism {
+            EncryptionMechanism::SeipdV1(sym) => SymKeyEncryptedSessionKey::encrypt_v4(
+                || pass,
+                &session_key,
+                StringToKey::new_default(&mut rng),
+                sym,
+            )?,
+            EncryptionMechanism::SeipdV2(aead, sym) => {
+                // "If much less memory is available, a uniformly safe option is Argon2id with
+                // t=3 iterations, p=4 lanes, m=2^(16) (64 MiB of RAM),
+                // 128-bit salt, and 256-bit tag size. This is the SECOND RECOMMENDED option."
+
+                // FIXME: move these settings up to rPGP, as part of a convenience constructor?
+                let s2k = StringToKey::new_argon2(&mut rng, 3, 4, 16);
+
+                SymKeyEncryptedSessionKey::encrypt_v6(
+                    &mut rng,
+                    || pass,
+                    &session_key,
+                    s2k,
+                    sym,
+                    aead,
+                )?
+            }
+        };
 
         esk.push(Esk::SymKeyEncryptedSessionKey(skesk));
     }
@@ -427,12 +430,13 @@ pub fn encrypt(
     let lit = LiteralData::from_bytes((&[]).into(), &data);
 
     // 5. Maybe sign the message.
-    let hash_algo = match hash_algo {
-        None => HashAlgorithm::default(),
-        Some(hash_algo) => *hash_algo,
-    };
+    let hash_algo = hash_algo.unwrap_or_default();
 
     let msg = if !signers.is_empty() {
+        // FIXME: set Intended Recipient Fingerprint? (at least for v6)
+        // "The OpenPGP Key fingerprint of the intended recipient primary key"
+        // "When generating this subpacket in a version 6 signature, it SHOULD be marked as critical."
+
         let mut packets = vec![];
         packets.push(Packet::from(lit.clone()));
 
@@ -467,22 +471,41 @@ pub fn encrypt(
             }
         }
 
-        Message::from_packets(packets.into_iter().map(Ok).peekable())
-            .next()
-            .expect("should be a message")
-            .expect("FIXME")
+        if let Some(Ok(msg)) = Message::from_packets(packets.into_iter().map(Ok).peekable()).next()
+        {
+            msg
+        } else {
+            // This shouldn't happen, if we construct a reasonable "packets"
+            return Err(Error::Message(
+                "Failed to construct Message from packets".to_string(),
+            ));
+        }
     } else {
         // we're not signing
         Message::Literal(lit)
     };
 
     // 6. Symmetrically Encrypt the message to the session key.
-    let edata = Edata::SymEncryptedProtectedData(SymEncryptedProtectedData::encrypt_with_rng(
-        &mut rng,
-        symmetric_algo,
-        &session_key,
-        &msg.to_bytes()?,
-    )?);
+    let edata = match mechanism {
+        EncryptionMechanism::SeipdV1(sym) => {
+            Edata::SymEncryptedProtectedData(SymEncryptedProtectedData::encrypt_seipdv1(
+                &mut rng,
+                sym,
+                &session_key,
+                &msg.to_bytes()?,
+            )?)
+        }
+        EncryptionMechanism::SeipdV2(aead, sym) => {
+            Edata::SymEncryptedProtectedData(SymEncryptedProtectedData::encrypt_seipdv2(
+                &mut rng,
+                sym,
+                aead,
+                crate::policy::AEAD_CHUNK_SIZE,
+                &session_key,
+                &msg.to_bytes()?,
+            )?)
+        }
+    };
 
     // 7. Put together encrypted message
     let msg = Message::Encrypted { esk, edata };
